"""This subpackage contains the classes for defining the target in SRIM simulations."""

from irradiapy.srimpy.target.element import Element
from irradiapy.srimpy.target.layer import Layer
from irradiapy.srimpy.target.target import Target
